﻿using Coscine.ApiClient.Core.Model;
using System;
using System.Collections.Generic;
using System.Reflection;
using System.Threading.Tasks;

namespace Coscine.ApiClient;

/// <summary>
/// Provides utility methods for handling web requests that return paged responses.
/// </summary>
public class RequestUtil
{
    /// <summary>
    /// Wraps a paged request into an asynchronous enumerable. This method utilizes reflection to handle various types of paged responses since they do not implement a common interface.
    /// </summary>
    /// <remarks>
    /// The method is generic and uses reflection to access 'Data', 'Pagination', and 'IsSuccess' properties of the paged response.
    /// This approach is not ideal in terms of performance and type safety, but it allows for handling different paged response types.
    /// </remarks>
    /// <typeparam name="S">The type of the paged response expected from the retrieval function.</typeparam>
    /// <typeparam name="T">The type of the data transfer object (DTO) contained within the paged response.</typeparam>
    /// <param name="RetrieveRequest">
    /// A function that takes an integer representing the current page number and returns a Task that resolves to a paged response of type <typeparamref name="S"/>.
    /// </param>
    /// <returns>
    /// A task representing the asynchronous operation. The task result contains an <see cref="IEnumerable{T}"/> that iterates over the retrieved DTOs.
    /// </returns>
    public static async Task<IEnumerable<T>> WrapPagedRequest<S, T>(
        Func<int, Task<S>> RetrieveRequest
    )
    {
        var returnArray = new List<T>();
        var hasNext = true;
        var currentPageNumber = 1;
        while (hasNext)
        {
            var iteration = await RetrieveRequest(currentPageNumber);

            PropertyInfo? dataProperty = iteration?.GetType().GetProperty("Data");
            PropertyInfo? paginationProperty = iteration?.GetType().GetProperty("Pagination");
            PropertyInfo? isSuccessProperty = iteration?.GetType().GetProperty("IsSuccess");

            if (isSuccessProperty != null &&
                (bool?)isSuccessProperty.GetValue(iteration) == true &&
                dataProperty != null &&
                paginationProperty != null)
            {
                var data = (List<T>?)dataProperty.GetValue(iteration);

                if (data != null)
                {
                    returnArray.AddRange(data);
                }

                var pagination = paginationProperty.GetValue(iteration);
                hasNext = pagination != null && ((Pagination)pagination).HasNext;
            }
            else
            {
                break;
            }

            currentPageNumber++;
        }

        return returnArray;
    }
}
